$!-----------------------------------------------------------------------------
$! BUILD_HTTPD.COM
$!
$! P1 ... P4  "UCX" or "NETLIB" or "TCPWARE" to specify IP
$! P1 ... P4  "SSL" to build with OpenSSL
$! P1 ... P4  "BUILD" to compile then link
$! P1 ... P4  "LINK" to link only
$! P1 ... P4  "DBUG" compiles with debug statements
$! P1 ... P4  "JUST:module" to compile just that module
$!
$! To specify a non-WASD-default location for the SSL compile-time include
$! files define a logical name BUILD_HTTPD_SSL to point to the parent
$! directory containing the OpenSSL distribution.  For example:
$!
$!   DEFINE /PROCESS BUILD_HTTPD_SSL_DIR LOCAL:[OPENSSL-0_9_3]
$!
$! Similarly for the link-time object libraries use the logical names provided
$! in the following example:
$!
$!   DEFINE /PROCESS BUILD_HTTPD_LIBSSL -
$!                   LOCAL:[OPENSSL-0_9_3.VAX.EXE.SSL]LIBSSL.OLB
$!   DEFINE /PROCESS BUILD_HTTPD_LIBCRYPTO -
$!                   LOCAL:[OPENSSL-0_9_3.VAX.EXE.CRYPTO]LIBCRYPTO.OLB
$!
$! 08-SEP-99  MGD  support for OpenSSL v0.9.4
$! 02-APR-99  MGD  generalize for OpenSSL
$! 24-DEC-98  MGD  HTTPd version 6,
$!                 add proxy modules,
$!                 suppress informational messages (DECC 5.7)
$! 03-JUL-98  MGD  SSLeay v0.9.0b,
$!                 build for TCPware in UCX compatibility mode
$! 14-MAY-98  MGD  add defines for __VMS_VER, __CRTL_VER, __VMS_V6_SOURCE
$!                 (ensure v7.1 compilations can be executed on v6.n),
$!                 distributing as .OBJ ... need UCX and NETLIB object files
$! 07-JAN-98  MGD  HTTPd version 5
$!                 added SSL (SSLeay) build support
$!                 removed optimization, seemed to be inducing problems
$! 07-SEP-97  MGD  modified to support DBUG and to retain NETLIB objects
$! 23-JUL-97  MGD  MultiNet dropped in favour of NETLIB 2.1
$! 01-FEB-97  MGD  HTTPd version 4
$! 06-JAN-96  MGD  build for either UCX or MultiNet
$! 01-DEC-95  MGD  HTTPd version 3
$! 24-MAY-95  MGD  build for AXP systems
$! 20-DEC-94  MGD  multi-threaded server
$! 20-JUN-94  MGD  single-threaded server
$!-----------------------------------------------------------------------------
$!
$ IP = ""
$ IF P1 .EQS. "NETLIB" .OR. P2 .EQS. "NETLIB" .OR. -
     P3 .EQS. "NETLIB" .OR. P4 .EQS. "NETLIB"
$ THEN
$    IP = "NETLIB"
$    IP_EXE = "_NETLIB"
$    IP_OBJ = "_NETLIB"
$ ENDIF
$ IF P1 .EQS. "TCPWARE" .OR. P2 .EQS. "TCPWARE" .OR. -
     P3 .EQS. "TCPWARE" .OR. P4 .EQS. "TCPWARE"
$ THEN
$!   (TCPware in UCX compatibility mode)
$    IP = "TCPWARE"
$    IP_EXE = ""
$    IP_OBJ = ""
$ ENDIF
$ IF IP .EQS. ""
$ THEN
$    IP = "UCX"
$    IP_EXE = ""
$    IP_OBJ = ""
$ ENDIF
$!
$ JUST = ""
$ IF F$ELEMENT(0,":",P1) .EQS. "JUST" THEN JUST = F$ELEMENT(1,":",P1)
$ IF F$ELEMENT(0,":",P2) .EQS. "JUST" THEN JUST = F$ELEMENT(1,":",P2)
$ IF F$ELEMENT(0,":",P3) .EQS. "JUST" THEN JUST = F$ELEMENT(1,":",P3)
$ IF F$ELEMENT(0,":",P4) .EQS. "JUST" THEN JUST = F$ELEMENT(1,":",P4)
$!
$ IF P1 .EQS. "SSL" .OR. P2 .EQS. "SSL" .OR. -
     P3 .EQS. "SSL" .OR. P4 .EQS. "SSL" .OR. -
     JUST .EQS. "SSL"
$ THEN
$    SESOLA = "SESOLA"
$    SESOLA_EXE = "_SSL"
$    SESOLA_OBJ = "_SSL"
$    SESOLA_INCLUDED = "with SSL"
$!
$    SSL_DIR = F$TRNLNM("BUILD_HTTPD_SSL_DIR")
$    IF SSL_DIR .NES. ""
$    THEN
$       TYPE SYS$INPUT

  ****************************************
  *  USER-DEFINED "BUILD_HTTPD_SSL_DIR"  *
  ****************************************

$       SSL_DIR = SSL_DIR - "]"
$    ENDIF
$!
$    IF SSL_DIR .EQS. ""
$    THEN
$!      (default WASD locations)
$       IF F$SEARCH("HT_ROOT:[SRC]SSLEAY-0_8_1.DIR") .NES. "" -
           THEN SSL_DIR = "HT_ROOT:[SRC.SSLEAY-0_8_1"
$       IF F$SEARCH("HT_ROOT:[SRC]SSLEAY-0_9_0B.DIR") .NES. "" -
           THEN SSL_DIR = "HT_ROOT:[SRC.SSLEAY-0_9_0B"
$       IF F$SEARCH("HT_ROOT:[SRC]OPENSSL-0_9_3.DIR") .NES. "" -
           THEN SSL_DIR = "HT_ROOT:[SRC.OPENSSL-0_9_3"
$       IF F$SEARCH("HT_ROOT:[SRC]OPENSSL-0_9_4.DIR") .NES. "" -
           THEN SSL_DIR = "HT_ROOT:[SRC.OPENSSL-0_9_4"
$    ENDIF
$!
$    OPENSSL = 0
$    IF F$LOCATE("OPENSSL",SSL_DIR) .LT. F$LENGTH(SSL_DIR) THEN OPENSSL = 1
$!
$    IF F$TYPE(SSL_DIR) .EQS. ""
$    THEN
$       TYPE SYS$INPUT

  *******************************
  *  SSL directory not located  *
  *******************************

$       EXIT
$    ENDIF
$!
$    LIBSSL_LIBRARY = F$TRNLNM("BUILD_HTTPD_LIBSSL")
$    LIBCRYPTO_LIBRARY = F$TRNLNM("BUILD_HTTPD_LIBCRYPTO")
$!
$    IF LIBSSL_LIBRARY .NES. ""
$    THEN
$       TYPE SYS$INPUT

  **********************************************
  *  USER-DEFINED "BUILD_HTTPD_LIBSSL/CRYPTO"  *
  **********************************************

$    ELSE
$!
$       IF F$EDIT(F$GETSYI("ARCH_NAME"),"UPCASE") .EQS. "VAX"
$       THEN
$          IF (OPENSSL)
$          THEN
$             LIBSSL_LIBRARY = "''SSL_DIR'.VAX.EXE.SSL]LIBSSL.OLB"
$             LIBCRYPTO_LIBRARY = "''SSL_DIR'.VAX.EXE.CRYPTO]LIBCRYPTO.OLB"
$          ELSE
$             SESOLA_OBJ = SESOLA_OBJ + "EAY"
$             LIBSSL_LIBRARY = "''SSL_DIR'.SSL]SSL-VAX.OLB"
$             LIBCRYPTO_LIBRARY = "''SSL_DIR'.CRYPTO]CRYPTO-VAX.OLB"
$          ENDIF
$       ELSE
$          IF (OPENSSL)
$          THEN
$             LIBSSL_LIBRARY = "''SSL_DIR'.AXP.EXE.SSL]LIBSSL.OLB"
$             LIBCRYPTO_LIBRARY = "''SSL_DIR'.AXP.EXE.CRYPTO]LIBCRYPTO.OLB"
$          ELSE
$             SESOLA_OBJ = SESOLA_OBJ + "EAY"
$             LIBSSL_LIBRARY = "''SSL_DIR'.SSL]SSL-AXP.OLB"
$             LIBCRYPTO_LIBRARY = "''SSL_DIR'.CRYPTO]CRYPTO-AXP.OLB"
$          ENDIF
$       ENDIF
$!
$    ENDIF
$!
$    IF (OPENSSL)
$       THEN SESOLA_OPENSSL = "SESOLA_OPENSSL=1"
$       ELSE SESOLA_OPENSSL = ""
$    ENDIF
$!
$    SSL_LIBRARY = "''LIBSSL_LIBRARY'/LIBRARY,''LIBCRYPTO_LIBRARY'/LIBRARY"
$!
$ ELSE
$!
$    SESOLA = ""
$    SESOLA_EXE = ""
$    SESOLA_OBJ = ""
$    SESOLA_INCLUDED = "without SSL"
$    SESOLA_OPENSSL = ""
$    OPENSSL = 0
$    SSL_DIR = ""
$    SSL_LIBRARY = ""
$!
$ ENDIF
$!
$ IF P1 .EQS. "DBUG" .OR. P2 .EQS. "DBUG" .OR. -
     P3 .EQS. "DBUG" .OR. P4 .EQS. "DBUG"
$ THEN
$    DBUG = "DBUG"
$ ELSE
$    DBUG = ""
$ ENDIF
$!
$ DEFINES = "_VMS_V6_SOURCE=1,__VMS_VER=60000000,__CRTL_VER=6000000"
$ IF IP .NES. "" THEN DEFINES = DEFINES + ","
$ DEFINES = DEFINES + "IP_" + IP
$ IF DBUG .NES. "" THEN DEFINES = DEFINES + ","
$ DEFINES = DEFINES + DBUG
$ IF SESOLA .NES. "" THEN DEFINES = DEFINES + ","
$ DEFINES = DEFINES + SESOLA
$ IF SESOLA_OPENSSL .NES. "" THEN DEFINES = DEFINES + ","
$ DEFINES = DEFINES + SESOLA_OPENSSL
$!
$ WRITE SYS$OUTPUT "Building for ""''F$GETSYI("ARCH_NAME")'"" architecture" + -
                   ", ""''IP'"" TCP/IP'', ''SESOLA_INCLUDED'."
$!
$ IF F$EDIT(F$GETSYI("ARCH_NAME"),"UPCASE") .EQS. "VAX"
$ THEN
$    OBJ_DIR = "[.OBJ_VAX]"
$    IF F$SEARCH("OBJ_VAX.DIR") .EQS. "" THEN CREATE /DIR 'OBJ_DIR'
$    CC_OPTIONS = "/DECC /STAND=VAXC /OPTIMIZE /DEF=(''DEFINES') " +-
                  "/NODEBUG /WARNING=NOINFORM"
$    LINK_OPTIONS = "[]HTTPD_VAX''SESOLA_OBJ'''IP_EXE'.OPT /OPTIONS"
$ ELSE
$    OBJ_DIR = "[.OBJ_AXP]"
$    IF F$SEARCH("OBJ_AXP.DIR") .EQS. "" THEN CREATE /DIR 'OBJ_DIR'
$    CC_OPTIONS = "/STAND=VAXC /OPTIMIZE /DEF=(''DEFINES') /NODEBUG"
$    LINK_OPTIONS = "[]HTTPD_AXP''SESOLA_OBJ'''IP_EXE'.OPT /OPTIONS"
$ ENDIF
$!
$ IF P1 .EQS. "LINKMAP" .OR. P2 .EQS. "LINKMAP" .OR. -
     P3 .EQS. "LINKMAP" .OR. P4 .EQS. "LINKMAP"
$ THEN
$    LINKMAP = " /MAP=HTTPD''SESOLA_EXE'''IP_EXE'.MAP /FULL"
$    IF P1 .EQS. "LINKMAP" THEN P1 = "LINK"
$    IF P2 .EQS. "LINKMAP" THEN P2 = "LINK"
$    IF P3 .EQS. "LINKMAP" THEN P3 = "LINK"
$    IF P4 .EQS. "LINKMAP" THEN P4 = "LINK"
$ ELSE
$    LINKMAP = ""
$ ENDIF
$!
$ IF P1 .EQS. "LINK" .OR. P2 .EQS. "LINK" .OR. -
     P3 .EQS. "LINK" .OR. P4 .EQS. "LINK" THEN GOTO LINK
$!
$! ***********
$! * COMPILE *
$! ***********
$!
$ SET NOON
$ SET VERIFY
$ IF JUST .NES. "" THEN GOTO 'JUST'
$ ADMIN:   CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'Admin'IP_OBJ'    Admin
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ AUTH:    CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'Auth'IP_OBJ'     Auth
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ BASIC:   CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'Basic'IP_OBJ'    Basic
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ CACHE:   CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'Cache'IP_OBJ'    Cache
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ CGI:     CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'CGI'IP_OBJ'      CGI
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ CLI:     CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'CLI'IP_OBJ'      CLI
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ CONFIG:  CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'Config'IP_OBJ'   Config
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ CONTROL: CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'Control'IP_OBJ'  Control
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ DCL:     CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'Dcl'IP_OBJ'      Dcl
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ DECNET:  CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'DECnet'IP_OBJ'      DECnet
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ DESCR:   CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'Descr'IP_OBJ'    Descr
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ DIR:     CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'Dir'IP_OBJ'      Dir
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ DIGEST:  CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'Digest'IP_OBJ'   Digest
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ ERROR:   CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'Error'IP_OBJ'    Error
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ FILE:    CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'File'IP_OBJ'     File
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ GRAPH:   CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'Graph'IP_OBJ'    Graph
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ HTADMIN: CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'HTAdmin'IP_OBJ'  HTAdmin
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ HTTPD:   CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'HTTPd'IP_OBJ'    HTTPd
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ ISMAP:   CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'IsMap'IP_OBJ'    IsMap
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ LOGGING: CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'Logging'IP_OBJ'  Logging
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ MAPURL:  CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'MapUrl'IP_OBJ'   MapUrl
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ MD5:     CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'MD5'IP_OBJ'      MD5
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ MENU:    CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'Menu'IP_OBJ'     Menu
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ MSG:     CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'Msg'IP_OBJ'      Msg
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ NET:     CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'Net'IP_OBJ'      Net
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ PROXY:   CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'Proxy'IP_OBJ'    Proxy
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ PROXYCACHE:  CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'ProxyCache'IP_OBJ'    ProxyCache
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ PROXYMAINT:  CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'ProxyMaint'IP_OBJ'    ProxyMaint
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ PUT:     CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'Put'IP_OBJ'      Put
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ REQUEST: CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'Request'IP_OBJ'  Request
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ SSI:     CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'SSI'IP_OBJ'      SSI
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ SESOLA:
$ IF SESOLA .EQS. ""
$ THEN
$    CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'SeSoLa'SESOLA_OBJ''IP_OBJ' SeSoLa
$ ELSE
$    IF (OPENSSL)
$    THEN
$       IF F$SEARCH("''SSL_DIR'.INCLUDE.OPENSSL]*.H") .NES. "" 
$       THEN
$          DEFINE /NOLOG /PROCESS /USER OPENSSL 'SSL_DIR'.INCLUDE.OPENSSL]
$       ELSE
$          DEFINE /NOLOG /PROCESS /USER OPENSSL 'SSL_DIR'.INCLUDE]
$       ENDIF
$    ELSE
$!      (WASD SSLEAY-0_8_1 and SSLEAY-0_9_0B)
$       DEFINE /USER DECC$USER_INCLUDE 'SSL_DIR'...]
$    ENDIF
$    CC 'CC_OPTIONS' /NOWARN /OBJECT='OBJ_DIR'SeSoLa'SESOLA_OBJ''IP_OBJ' SeSoLa
$ ENDIF
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ STMLF:   CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'StmLF'IP_OBJ'    StmLF
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ SUPPORT: CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'Support'IP_OBJ'  Support
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ UPD:     CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'Upd'IP_OBJ'      Upd
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ VM:      CC 'CC_OPTIONS' /OBJECT='OBJ_DIR'VM'IP_OBJ'       VM
$ IF JUST .NES. "" THEN GOTO ENDCOMPILE
$ ENDCOMPILE:
$! 'F$VERIFY(0)

$ IF P1 .EQS. "BUILD" .OR. P2 .EQS. "BUILD" .OR. -
     P3 .EQS. "BUILD" .OR. P4 .EQS. "BUILD" -
     THEN GOTO LINK
$ IF JUST .NES. "" THEN EXIT
$!
$! ********
$! * LINK *
$! ********
$!
$ LINK:
$ SET ON
$ SET VERIFY
$!
$ IF SSL_LIBRARY .NES. "" THEN LINK_OPTIONS = LINK_OPTIONS + ", "
$ LINK'LINKMAP' /NOTRACE /EXECUTABLE=HT_EXE:HTTPD'SESOLA_EXE''IP_EXE'.EXE -
  'LINK_OPTIONS' -
  'SSL_LIBRARY'
$!
$ PURGE /LOG 'OBJ_DIR' /KEEP=3
$ SET PROTECTION=W:E HT_EXE:HTTPD'SESOLA_EXE''IP_EXE'.EXE;*
$!'F$VERIFY(0)
$!-----------------------------------------------------------------------------
