/*****************************************************************************
/*
                                  WASD.h

Main header file for WASD VMS HTTP daemon.


VERSION HISTORY
---------------
16-OCT-99  MGD  v6.0.3, DEC TCP/IP v5.n compatibility definitions
10-JAN-99  MGD  v6.0.0, proxy serving
19-SEP-98  MGD  v5.3.0, disperse to module headers :^)
14-APR-98  MGD  v5.0.1, XSSI extensions
07-JAN-98  MGD  v5.0.0, HTTPd version 5 (Secure Sockets Layer, using SSLeay)
05-OCT-97  MGD  v4.5.0, file cache
06-SEP-97  MGD  v4.4.0, multiple server host names and ports
08-JUN-97  MGD  v4.2.0, CGIplus
01-FEB-97  MGD  v4.0.0, HTTPd version 4
01-AUG-96  MGD  v3.3.0, quite a few changes since 3.0.n
01-DEC-95  MGD  v3.0.0, HTTPd version 3
20-DEC-94  MGD  v2.0.0, multi-threaded version
20-JUN-94  MGD  v1.0.0, single-threaded version
*/
/*****************************************************************************/

#ifndef WASD_H_LOADED
#define WASD_H_LOADED 1

#ifndef IP_NONE
/* if not NETLIB then UCX! */
#ifndef IP_NETLIB
#  undef IP_UCX
#  define IP_UCX 1
#endif
#endif

#include <descrip.h>
#include <rms.h>
#include <ssdef.h>
#include <stsdef.h>
#include <rmsdef.h>

/* Internet-related header files */
#ifdef IP_UCX

#   include <socket.h>
#   include <in.h>
#   include <netdb.h>
#   include <inet.h>
#   ifdef IP_TCPWARE
#      include "tcpware_include:ucx$inetdef.h"
#   else
#      include <ucx$inetdef.h>
#   endif

    /* for UCX 4.n - DEC TCP/IP 5.n compatibility */
#   ifndef UCX$C_FULL_DUPLEX_CLOSE
#      ifndef TCPIP$C_FULL_DUPLEX_CLOSE
          /* provide this for older versions of UCX (just which ones?) */
#         define UCX$C_FULL_DUPLEX_CLOSE 8192
#      else
#         define UCX$C_FULL_DUPLEX_CLOSE TCPIP$C_FULL_DUPLEX_CLOSE
#      endif
#   endif
#   ifndef UCX$C_REUSEADDR
#      define UCX$C_REUSEADDR TCPIP$C_REUSEADDR
#   endif
#   ifndef UCX$C_SOCKOPT
#      define  UCX$C_SOCKOPT TCPIP$C_SOCKOPT
#   endif
#   ifndef UCX$C_TCP
#      define UCX$C_TCP TCPIP$C_TCP
#   endif
#   ifndef UCX$C_AF_INET
#      define UCX$C_AF_INET TCPIP$C_AF_INET
#   endif

#endif

#ifdef IP_NETLIB
#   include "netlibdef.h"
#endif

/* Alpha NO MEMBER ALIGNMENT unless otherwise specified */
#ifdef __ALPHA
#   pragma nomember_alignment
#endif

/**********/
/* macros */
/**********/

#define boolean int
#define true 1
#define false 0

#define VMSok(x) ((x) & STS$M_SUCCESS)
#define VMSnok(x) (!((x) & STS$M_SUCCESS))

#define FI_LI WASD_MODULE, __LINE__

#define CONFIG_AUTH_FILE_NAME  "HTTPD$AUTH"
#define CONFIG_FILE_NAME       "HTTPD$CONFIG"
#define CONFIG_MAP_FILE_NAME   "HTTPD$MAP"
#define CONFIG_MSG_FILE_NAME   "HTTPD$MSG"

/* this value must also include connections in a "keep-alive" state */
#define DEFAULT_CONCURRENT_CONNECT_MAX    50

#define DEFAULT_TIMEOUT_INPUT_MINUTES      1
#define DEFAULT_TIMEOUT_OUTPUT_MINUTES    10
#define DEFAULT_TIMEOUT_KEEPALIVE_SECONDS  0

/* helps ensure a client doesn't get into an infinite loop */
#define KEEPALIVE_MAX 100

#define DEFAULT_DCL_SYSOUTPUT_SIZE    4096
#define DEFAULT_FILE_BUFFER_SIZE      1024
#define DEFAULT_NET_READ_BUFFER_SIZE  2048
#define DEFAULT_OUTPUT_BUFFER_SIZE    4096
#define MAX_REQUEST_HEADER            8192

/* control pre-expired responses */
#define PRE_EXPIRE_ADMIN            1
#define PRE_EXPIRE_DELETE_ON_CLOSE  1
#define PRE_EXPIRE_INDEX_OF         0
#define PRE_EXPIRE_MAPPING_RULES    0
#define PRE_EXPIRE_MENU             1
#define PRE_EXPIRE_PUT              1
#define PRE_EXPIRE_SSI              0
#define PRE_EXPIRE_UPD              1
#define PRE_EXPIRE_UPD_TREE_OF      1
#define PRE_EXPIRE_WATCH            0

/* versions of the HTTP protocol */
#define HTTP_VERSION_NOT_SET   0
#define HTTP_VERSION_0_9      09
#define HTTP_VERSION_1_0      10
#define HTTP_VERSION_1_1      11
#define HTTP_VERSION_UNKNOWN  -1

/* constants used to identify HTTP methods */
#define HTTP_METHOD_DELETE   0x01
#define HTTP_METHOD_GET      0x02
#define HTTP_METHOD_HEAD     0x04
#define HTTP_METHOD_POST     0x08
#define HTTP_METHOD_PUT      0x10
#define HTTP_METHOD_CONNECT  0x20

/* constants for controlling logging */
#define LOGGING_BEGIN  1
#define LOGGING_END    2
#define LOGGING_OPEN   3
#define LOGGING_CLOSE  4
#define LOGGING_FLUSH  5
#define LOGGING_ENTRY  6

/* constants used by HTTPd supervisor timer */
#define TIMER_INPUT             1
#define TIMER_OUTPUT            2
#define TIMER_KEEPALIVE         3

/* constants used by the description module */
#define DESCRIPTION_ALL         0x01
#define DESCRIPTION_TEXT_PLAIN  0x02
#define DESCRIPTION_TEXT_HTML   0x04
/* this is returned in the first character the file type contains none */
#define DESCRIPTION_IMPOSSIBLE  0xff

/* default directory listing layout */
#define DEFAULT_DIR_LAYOUT "I__L__R__S:b__D"

/* "http:", "https:" or "connect:" service and request */
#define SCHEME_HTTP     1
#define SCHEME_HTTPS    2
#define SCHEME_CONNECT  3

/* maximum Kbytes (1024) that can be in a body to be POSTed or PUTted */
#define DEFAULT_PUT_KBYTES_MAX 250
/* number of versions retained for files POSTed or PUTed */
#define DEFAULT_PUT_VERSION_LIMIT 3

/* header fields for HTTP/1.0 keep alive response */
#define DEFAULT_KEEPALIVE_HTTP_HEADER \
        "Connection: Keep-Alive\r\nKeep-Alive:\r\n"

/* if true, stops HTTPd server account with SYSPRV PUT(etc)ing - DANGEROUS */
#define PUT_DISALLOW_SYSPRV 0

#define CONFIG_HOME_PAGES_MAX      16
#define CONFIG_README_FILES_MAX    16
#define CONFIG_SCRIPT_RUNTIME_MAX  16

#define SCRIPT_NAME_SIZE  128
#define METHOD_NAME_SIZE   16

/* W:RE,G:RE,O:RWED,S:RWED */
#define PUT_DEFAULT_FILE_PROTECTION 0xaa00

/* number of times a request can internally redirect */
#define REQUEST_REDIRECTION_MAX 5

/*************/
/* durations */
/*************/

/* 1000 = 1mS, 100 = 100uS, 10 = 10Us, 1 = 1uS (for really fast systems :^) */
#define DURATION_UNITS_USECS 1000

/* durations are to be in units of 0.1 milliseconds */
#define DURATION_UNITS_USECS 100

#if DURATION_UNITS_USECS == 1
#  define DURATION_DECIMAL_PLACES 6
#elif DURATION_UNITS_USECS == 10
#  define DURATION_DECIMAL_PLACES 5
#elif DURATION_UNITS_USECS == 100
#  define DURATION_DECIMAL_PLACES 4
#else
   /* defaults to milliseconds */
#  undef DURATION_UNITS_USECS
#  define DURATION_UNITS_USECS 1000
#  define DURATION_DECIMAL_PLACES 3
#endif

#define USECS_IN_A_SECOND 1000000

/*********************/
/* functional macros */
/*********************/

/* is it linear-white-space? (avoids the call overhead of an isspace()) */
#define ISLWS(c) (c == ' ' || c == '\t')

/* end of CR or LF terminated line? */
#define EOL(c) (c == '\r' || c == '\n' || !*cptr)
#define NOTEOL(c) (c != '\r' && c != '\n' && *cptr)

/*************************************************/
/* macros for development timing in microseconds */
/*************************************************/

/* 0..9 timers available */

#define USEC_TIMER_STORAGE \
   unsigned long  uSecSubx2 = 2, \
                  uSecEdiv10 = 10; \
   unsigned long  uSecResult, \
                  uSecDuration, \
                  uSecRemainder; \
   unsigned long  uSecBeginQuad [10][2], \
                  uSecEndQuad [10][2];

#define USEC_TIMER_STORAGE_EXTERN \
   extern unsigned long  uSecSubx2, \
                         uSecEdiv10; \
   extern unsigned long  uSecResult, \
                         uSecDuration, \
                         uSecRemainder; \
   extern unsigned long  uSecBeginQuad [10][2], \
                         uSecEndQuad [10][2];

#define USEC_TIMER_BEGIN(IDX) \
{ \
   sys$gettim (&uSecBeginQuad[IDX]); \
}

#define USEC_TIMER_END(IDX,WHERE) \
{ \
   sys$gettim (&uSecEndQuad[IDX]); \
   lib$subx (&uSecEndQuad[IDX], &uSecBeginQuad[IDX], &uSecResult, &uSecSubx2); \
   lib$ediv (&uSecEdiv10, &uSecResult, &uSecDuration, &uSecRemainder); \
   fprintf (stdout, "%s uS: %d\n", WHERE, uSecDuration); \
}

/*********************/
/* admin menu macros */
/*********************/

#define INTERNAL_SCRIPT_ECHO   "/echo"
#define INTERNAL_SCRIPT_TREE   "/tree"
#define INTERNAL_SCRIPT_UPD    "/upd"
#define INTERNAL_SCRIPT_WHERE  "/where"
#define INTERNAL_SCRIPT_XRAY   "/xray"

#define INTERNAL_PASSWORD_CHANGE  "/httpd/-/change/"

#define INTERNAL_ADMIN  "/httpd/-/admin/"

#define INTERNAL_CONTROL              INTERNAL_ADMIN "control/"
#define INTERNAL_CONTROL_AUTH_LOAD    INTERNAL_CONTROL "auth/load"
#define INTERNAL_CONTROL_AUTH_PURGE   INTERNAL_CONTROL "auth/purge"
#define INTERNAL_CONTROL_CACHE_ON     INTERNAL_CONTROL "cache/on"
#define INTERNAL_CONTROL_CACHE_OFF    INTERNAL_CONTROL "cache/off"
#define INTERNAL_CONTROL_CACHE_PURGE  INTERNAL_CONTROL "cache/purge"
#define INTERNAL_CONTROL_DCL_PURGE    INTERNAL_CONTROL "DCL/purge"
#define INTERNAL_CONTROL_DCL_DELETE   INTERNAL_CONTROL "DCL/delete"
#define INTERNAL_CONTROL_DECNET_FORCE INTERNAL_CONTROL "DECnet/force"
#define INTERNAL_CONTROL_DECNET_PURGE INTERNAL_CONTROL "DECnet/purge"
#define INTERNAL_CONTROL_EXIT         INTERNAL_CONTROL "exit"
#define INTERNAL_CONTROL_EXIT_NOW     INTERNAL_CONTROL "exit/now"
#define INTERNAL_CONTROL_LOG_ON       INTERNAL_CONTROL "log/on"
#define INTERNAL_CONTROL_LOG_OFF      INTERNAL_CONTROL "log/off"
#define INTERNAL_CONTROL_LOG_FLUSH    INTERNAL_CONTROL "log/flush"
#define INTERNAL_CONTROL_MAPPING_LOAD INTERNAL_CONTROL "mapping/load"
#define INTERNAL_CONTROL_PROXY        INTERNAL_CONTROL "proxy"
#define INTERNAL_CONTROL_PROXY_ADJUST INTERNAL_CONTROL "proxy/adjust"
#define INTERNAL_CONTROL_RESTART      INTERNAL_CONTROL "restart"
#define INTERNAL_CONTROL_RESTART_NOW  INTERNAL_CONTROL "restart/now"
#define INTERNAL_CONTROL_ZERO         INTERNAL_CONTROL "accounting/zero"

#define INTERNAL_GRAPHIC           INTERNAL_ADMIN   "graphic/"
#define INTERNAL_GRAPHIC_ACTIVITY  INTERNAL_GRAPHIC "activity"

#define INTERNAL_REPORT            INTERNAL_ADMIN  "report/"
#define INTERNAL_REPORT_ACTIVITY   INTERNAL_REPORT "activity"
#define INTERNAL_REPORT_AUTH_USER  INTERNAL_REPORT "auth/user"
#define INTERNAL_REPORT_AUTH_PATHS INTERNAL_REPORT "auth/paths"
#define INTERNAL_REPORT_CACHE      INTERNAL_REPORT "cache"
#define INTERNAL_REPORT_CACHE_ENTRIES  INTERNAL_REPORT "cache/entries"
#define INTERNAL_REPORT_CONFIG     INTERNAL_REPORT "config"
#define INTERNAL_REPORT_DCL        INTERNAL_REPORT "dcl"
#define INTERNAL_REPORT_DECNET     INTERNAL_REPORT "DECnet"
#define INTERNAL_REPORT_MEMORY     INTERNAL_REPORT "memory"
#define INTERNAL_REPORT_MESSAGES   INTERNAL_REPORT "messages"
#define INTERNAL_REPORT_PROXY      INTERNAL_REPORT "proxy"
#define INTERNAL_REPORT_REQUEST    INTERNAL_REPORT "request"
#define INTERNAL_REPORT_REQUEST_HISTORY  INTERNAL_REPORT "request/history"
#define INTERNAL_REPORT_MAPPING    INTERNAL_REPORT "mapping"
#define INTERNAL_REPORT_SSL        INTERNAL_REPORT "SSL"
#define INTERNAL_REPORT_STATS      INTERNAL_REPORT "stats"
#define INTERNAL_REPORT_WATCH      INTERNAL_REPORT "WATCH"

#define INTERNAL_REVISE            INTERNAL_ADMIN  "revise/"
#define INTERNAL_REVISE_AUTH_PATHS INTERNAL_REVISE "auth/paths"
#define INTERNAL_REVISE_CONFIG     INTERNAL_REVISE "config"
#define INTERNAL_REVISE_HTA        INTERNAL_REVISE "hta/"
#define INTERNAL_REVISE_MESSAGES   INTERNAL_REVISE "messages"
#define INTERNAL_REVISE_MAPPING    INTERNAL_REVISE "rules"

/*******************/
/* data structures */
/*******************/

struct VmsItem
{
   unsigned short  buf_len;
   unsigned short  item;
   unsigned char   *buf_addr;
   unsigned short  *ret_len;
};

struct AnExitHandler
{
   unsigned long  Flink;
   unsigned long  HandlerAddress;
   unsigned long  ArgCount;
   unsigned long  ExitStatusPtr;
};

struct AnIOsb
{
   unsigned short  Status;
   unsigned short  Count;
   unsigned long  Unused;
};

struct MbxSenseIOsb
{
   unsigned short  Status;
   unsigned short  MessageCount;
   unsigned long  MessageBytes;
};

/*
   Store these structures naturally-aligned on AXP.
   Uses a bit more storage but makes access as efficient as possible.
*/

#ifdef __ALPHA
#   pragma member_alignment __save
#   pragma member_alignment
#endif

/* per-request memory heap */
struct HeapStruct
{
   struct HeapStruct  *PrevPtr;
   struct HeapStruct  *NextPtr;
   unsigned int  ChunkSize;
   /* the actual data follows immediately after the length */
};

/*
This structure is used to both store and to point at the various components
depending on the context.
*/

struct ContentTypeStruct
{
   struct ContentTypeStruct  *NextPtr,
                             *HashCollisionPtr;
   boolean  ContentTypeUnknown;
   char  *SuffixPtr;
   char  *ContentTypePtr;
   char  *AutoScriptNamePtr;
   char  *ContentDescriptionPtr;
   char  *IconPtr;
   /* string data is stored here */
};

/*
This structure is used to both store and to point at the various components
depending on the context.
*/

struct IconStruct
{
   struct IconStruct  *NextPtr,
                      *HashCollisionPtr;
   /* string data is stored here */
};

struct  ListEntryStruct
{
   struct ListEntryStruct  *PrevPtr;
   struct ListEntryStruct  *NextPtr;
};

struct  ListHeadStruct
{
   struct ListEntryStruct  *HeadPtr;
   struct ListEntryStruct  *TailPtr;
   int  EntryCount;
};

/* these macros operate on the list head */
#define LIST_HEAD(lp) ((char*)((struct ListHeadStruct*)lp)->HeadPtr)
#define LIST_TAIL(lp) ((char*)((struct ListHeadStruct*)lp)->TailPtr)
#define LIST_COUNT(lp) (((struct ListHeadStruct*)lp)->EntryCount)
#define LIST_EMPTY(lp) (!((struct ListHeadStruct*)lp)->EntryCount)

/* these macros operate on list entries */
#define LIST_DATA(lp) ((char*)((char*)lp+sizeof(struct ListEntryStruct)))
#define LIST_PREV(lp) (((struct ListEntryStruct*)lp)->PrevPtr)
#define LIST_NEXT(lp) (((struct ListEntryStruct*)lp)->NextPtr)
#define LIST_HAS_PREV(lp) ((((struct ListEntryStruct*)lp)->PrevPtr) != NULL)
#define LIST_HAS_NEXT(lp) ((((struct ListEntryStruct*)lp)->NextPtr) != NULL)

/**********************************/
/* request history data structure */
/**********************************/

/* 256 bytes (a somewhat arbitrary quantity) */
struct  RequestHistoryStruct
{
   struct  ListEntryStruct  HistoryListEntry;
   struct  ServiceStruct  *ServicePtr;
   unsigned long  BinaryTime [2];
   unsigned long  BytesRawRx,
                  BytesRawTx,
                  ResponseDuration;
   int  ResponseStatusCode;
   char  *RequestPtr;
   boolean  Truncated;
   char  ClientAndRequest [256-12-8-4-4-4-4-4-4];
};

/***************************/
/* configuration structure */
/***************************/

/* must be powers of two! */
#define  CONTENT_TYPE_HASH_TABLE_SIZE  128
#define  CONTENT_TYPE_HASH_TABLE_MASK  CONTENT_TYPE_HASH_TABLE_SIZE-1
#define  ICON_HASH_TABLE_SIZE  64
#define  ICON_HASH_TABLE_MASK  ICON_HASH_TABLE_SIZE-1

struct ConfigStruct
{
   boolean  AuthBasicEnabled,
            AuthDigestEnabled,
            CacheEnabled,
            DclFullRequest,
            DclSpawnAuthPriv,
            DirAccess,
            DirAccessSelective,
            DirNoImpliedWildcard,
            DirNoPrivIgnore,
            DirOwnerEnabled,
            DirPreExpired,
            DirReadMeBottom,
            DirReadMeTop,
            DirWildcardEnabled,
            DnsLookup,
            ErrorRecommend,
            ErrorSourceInfo,
            IncludeCommentedInfo,
            LoggingEnabled,
            LogPerService,
            MonitorEnabled,
            ProxyAddForwardedByEnabled,
            ProxyReportCacheLog,
            ProxyReportLog,
            ProxyServingEnabled,
            SsiEnabled,
            SsiAccessesEnabled,
            SsiExecEnabled;

   int  AcceptHostsLength,
        ActivityNumberOfDays,
        AuthCacheMinutes,
        AuthDigestNonceGetLifeTime,
        AuthDigestNoncePutLifeTime,
        AuthFailureLimit,
        AuthRevalidateUserMinutes,
        Busy,
        CacheEntriesMax,
        CacheChunkKBytes,
        CacheFileKBytesMax,
        CacheFrequentHits,
        CacheFrequentSeconds,
        CacheHashTableEntries,
        CacheTotalKBytesMax,
        CacheValidateSeconds,
        DclCgiPlusLifeTime,
        DclSubprocessHardLimit,
        DclSubprocessSoftLimit,
        DclZombieLifeTime,
        DECnetConnectListMax,
        DECnetReuseLifeTime,
        DirDescriptionLines,
        ErrorReportPathLength,
        HomePageCount,
        InputTimeoutMinutes,
        KeepAliveTimeoutSeconds,
        KeywordSearchScriptLength,
        LogExcludeHostsLength,
        OutputTimeoutMinutes,
        ProxyCacheDeviceMaxPercent,
        ProxyCacheDevicePurgePercent,
        ProxyCacheDeviceCheckMinutes,
        ProxyCacheFileKBytesMax,
        ProxyCacheRoutineHourOfDay,
        ProxyHostCachePurgeHours,
        ProxyHostLookupRetryCount,
        PutMaxKbytes,
        PutVersionLimit,
        ReadMeFileCount,
        RejectHostsLength,
        RequestHistory,
        ScriptRunTimeCount,
        ServerPort,
        ServiceLength,
        StreamLfConversionMaxKbytes;

   int  ScriptRunTimeLengthArray [CONFIG_SCRIPT_RUNTIME_MAX];

   struct ContentTypeStruct  
      *ContentTypeHashTable [CONTENT_TYPE_HASH_TABLE_SIZE];
   struct IconStruct *IconHashTable [ICON_HASH_TABLE_SIZE];

   char  CharsetDefault [32],
         DefaultDirLayout [32],
         ErrorReportPath [128],
         KeywordSearchScript [128],
         LogFileName [128],
         LogFormat [128],
         LogNaming [32],
         LogPeriod [32],
         ProxyCachePurgeList [64],
         ProxyCacheReloadList [64],
         ServiceNotFoundUrl [128];

   char  *AcceptHostsPtr,
         *DefaultPost,
         *LogExcludeHostsPtr,
         *RejectHostsPtr,
         *ServicePtr;

   char  *HomePageArray [CONFIG_HOME_PAGES_MAX],
         *ReadMeFileArray [CONFIG_README_FILES_MAX],
         *ScriptRunTimeArray [CONFIG_SCRIPT_RUNTIME_MAX];

   struct ContentTypeStruct  *ContentTypeListHeadPtr,
                             *ContentTypeListTailPtr;
   char  *ContentTypeDefaultPtr;
   int  ContentTypeStructOverhead;

   struct IconStruct  *IconListHeadPtr,
                      *IconListTailPtr;
   int  IconStructOverhead;

   /* only used while reading configuration from file */
   int  LineNumber,
        ProblemCount,
        ProblemReportLength;
   char  *ProblemReportPtr;
   char  DirectiveName [32],
         LoadFileName [256];
   unsigned long  LoadBinTime [2],
                  RevBinTime [2];

   /* only used when reporting or changing configuration */
   void  *RequestPtr;
};

/*************************/
/* per-server accounting */
/*************************/

/* as of v4.4 this structure is now > 255 bytes requiring two logicals */

struct AccountingStruct
{
   /* this is used to check for changes in HTTPd version requiring zeroing */
   unsigned long  AccountingVersion;
   /* this MUST be the first member (or so) to allow zeroing of the rest */
   unsigned long  ZeroedCount;

   unsigned long  AuthBasicCount,
                  AuthAuthorizedCount,
                  AuthDigestCount,
                  AuthHtDatabaseCount,
                  AuthNotAuthenticatedCount,
                  AuthNotAuthorizedCount,
                  AuthVmsCount,
                  CacheHitCount,
                  CacheHitNotModifiedCount,
                  CacheLoadCount,
                  ConnectAcceptedCount,
                  ConnectCount,
                  ConnectSslCount,
                  ConnectCurrent,
                  ConnectPeak,
                  ConnectRejectedCount,
                  ConnectTooBusyCount,
                  DclCgiPlusReusedCount,
                  DclSpawnCount,
                  DoAutoScriptCount,
                  DoCgiPlusScriptCount,
                  DoDclCommandCount,
                  DoDECnetCount,
                  DoDECnetReuseCount,
                  DoDECnetCgiCount,
                  DoDECnetCgiReuseCount,
                  DoDECnetOsuCount,
                  DoDECnetOsuReuseCount,
                  DoDirectoryCount,
                  DoIsMapCount,
                  DoScriptCount,
                  DoFileCount,
                  DoFileNotModifiedCount,
                  DoMenuCount,
                  DoPutCount,
                  DoServerAdminCount,
                  DoSsiCount,
                  DoUpdateCount,
                  LastExitStatus,
                  MethodDeleteCount,
                  MethodGetCount,
                  MethodHeadCount,
                  MethodPostCount,
                  MethodPutCount,
                  MethodTraceCount,
                  RedirectLocalCount,
                  RedirectRemoteCount,
                  RequestTotalCount,
                  RequestErrorCount,
                  RequestForbiddenCount,
                  RequestKeepAliveCount,
                  RequestKeepAliveMax,
                  RequestParseCount,
                  RequestPragmaNoCacheCount,
                  RequestProxyCount,
                  ResponseDurationCount,
                  ResponseDurationMax,
                  ResponseDurationMin,
                  StartupCount,
                  StreamLfConversionCount;

   unsigned long  QuadBytesRawRx [2],
                  QuadBytesRawTx [2],
                  QuadResponseDuration [2];

   unsigned long  ResponseStatusCodeCountArray [6];
};

#ifdef __ALPHA
#   pragma member_alignment __restore
#endif

/************************/
/* module include files */
/************************/

#ifndef ADMIN_H_LOADED
#   include "admin.h"
#endif
#ifndef AUTH_H_LOADED
#   include "auth.h"
#endif
#ifndef BASIC_H_LOADED
#   include "basic.h"
#endif
#ifndef CACHE_H_LOADED
#   include "cache.h"
#endif
#ifndef CGI_H_LOADED
#   include "cgi.h"
#endif
#ifndef CONFIG_H_LOADED
#   include "config.h"
#endif
#ifndef CONTROL_H_LOADED
#   include "control.h"
#endif
#ifndef DESCR_H_LOADED
#   include "descr.h"
#endif
#ifndef DCL_H_LOADED
#   include "dcl.h"
#endif
#ifndef DIR_H_LOADED
#   include "dir.h"
#endif
#ifndef DECNET_H_LOADED
#   include "decnet.h"
#endif
#ifndef ERROR_H_LOADED
#   include "error.h"
#endif
#ifndef FILE_H_LOADED
#   include "file.h"
#endif
#ifndef GRAPH_H_LOADED
#   include "graph.h"
#endif
#ifndef HTADMIN_H_LOADED
#   include "htadmin.h"
#endif
#ifndef HTTPD_H_LOADED
#   include "httpd.h"
#endif
#ifndef ISMAP_H_LOADED
#   include "ismap.h"
#endif
#ifndef MAPURL_H_LOADED
#   include "mapurl.h"
#endif
#ifndef MENU_H_LOADED
#   include "menu.h"
#endif
#ifndef MSG_H_LOADED
#   include "msg.h"
#endif
#ifndef NET_H_LOADED
#   include "net.h"
#endif
#ifndef PROXY_H_LOADED
#   include "proxy.h"
#endif
#ifndef PROXYCACHE_H_LOADED
#   include "proxycache.h"
#endif
#ifndef PROXYMAINT_H_LOADED
#   include "proxymaint.h"
#endif
#ifndef PUT_H_LOADED
#   include "put.h"
#endif
#ifndef REQUEST_H_LOADED
#   include "request.h"
#endif
#ifndef SSI_H_LOADED
#   include "ssi.h"
#endif
#ifndef SESOLA_H_LOADED
#   include "sesola.h"
#endif
#ifndef STMLF_H_LOADED
#   include "stmlf.h"
#endif
#ifndef SUPPORT_H_LOADED
#   include "support.h"
#endif
#ifndef UPD_H_LOADED
#   include "upd.h"
#endif
#ifndef VM_H_LOADED
#   include "vm.h"
#endif

/*****************************/
/* per-thread data structure */
/*****************************/

/*
   Store these structure naturally-aligned on AXP.
   Uses a bit more storage but makes access as efficient as possible.
*/

#ifdef __ALPHA
#   pragma member_alignment __save
#   pragma member_alignment
#endif

struct RequestStruct
{
   /***************************/
   /* general purpose storage */
   /***************************/

   /*
      Storage areas at the beginning of this structure, specifically
      above the field 'RetainAboveZeroBelow', are retained between
      successive requests during persistent connections.
      When a "keep-alive" connection is reinitialized ready for a
      potential new request all of this structure below (and including)
      that field are zeroed, effectively allowing a new request.
   */

   /* for maintaining the list of request structures */
   struct  ListEntryStruct  RequestListEntry;

   /* pointer to service structure of host/port etc of service connected to */
   struct ServiceStruct  *ServicePtr;

   int  /* counter for input timeout */
        TimerInputCount,
        /* counter for keep-alive timeout */
        TimerKeepAliveCount,
        /* counter for output timeout */
        TimerOutputCount;

   unsigned long  /* zone ID for request heap's virtual memory management */
                  VmHeapZoneId;

#ifdef IP_UCX
   unsigned short  /* channel to client's TCP/IP device */
                   ClientChannel;
#endif

#ifdef IP_NETLIB
   unsigned long  /* socket for client's TCP/IP device */
                  ClientNetLibSocket;
#endif

   boolean  /* when the socket has been explicitly shutdown (before close) */
            ClientSocketShutdown,
            /* timers expired and the client socket has been shut/closed */
            TimerTerminatedCount;

   int  /* should be a short, but */
        ClientPort;

   unsigned long  /* client's binary IP address */
                  ClientIpAddress;

   unsigned char  /* internet host/domain name of client */
                  ClientHostName [128],
                  /* numeric host/domain address of client */
                  ClientIpAddressString [32];

   int  /* for life of the thread */
        AdminWatchItem,
        /* count of how many times the TCP/IP connection is reused */
        KeepAliveCount,
        /* an error is being reported from a redirect path */
        ErrorReportStatusCode,
        /* avoid redirection loops by keeping track of the number */
        RedirectionCount;

   char  /* when redirected this points to any authentication realm */
         *ErrorReportAuthRealmDescriptPtr;

   /* make it easy on ourselves, make the SSL structure pointer typeless! */
   void  *SeSoLaPtr;
   /* "http:" or "https:" (SSL) */
   int  RequestScheme;
   /* pointer to either "http:" or "https:" */
   char  *RequestSchemeNamePtr;

   /* AST function pointers */
   void (*NetReadRawAstFunctionPtr)(struct RequestStruct*);
   void (*NetShutdownAstFunctionPtr)(struct RequestStruct*);
   void (*NetWriteRawAstFunctionPtr)(struct RequestStruct*);

   /* the rest of this structure can be zeroed with persistent connections */
   /**********************************/
   unsigned long  RetainAboveZeroBelow;
   /**********************************/

   boolean  /* this request is using the watch report facility */
            AdminWatchClient,
            /* cache is currently being loaded for this path/file */
            CacheLoading,
            /* */
            CgiAddNewline,
            /* */
            CgiBufferRecords,
            /* */
            CgiCheckedHeader,
            /* */
            CgiContentTypeText,
            /* */
            CgiScript,
            /* */
            CgiPlusScript,
            /* "temporary" file, delete it on close */
            DeleteOnClose,
            /* report any errors using the error report script */
            ErrorReportByRedirect,
            /* true if a mapping rule identifies this as a "CGIplus" script */
            IsCgiPlusScript,
            /* this request has a "Connection: Keep-Alive" token */
            KeepAliveRequest,
            /* the server has responded with "keep-alive" */
            KeepAliveResponse,
            /* this persistent connection has timed out */
            KeepAliveTimeout,
            /* escapes HTML-forbidden characters as sent to client */
            NetWriteBufferedEscapeHtml,
            /* response header has been sent to client */
            ResponseHeaderSent,
            /* response include an "Expires:" the same as "Last-Modified:" */
            ResponsePreExpired,
            /* "Pragma: no-cache" HTTP request header line encountered */
            PragmaNoCache,
            /* only authorize the script, not any path component */
            SetPathAuthorizeOnce,
            /* rule mapping, SET the path to be pre-expired */
            SetPathExpired,
            /* rule mapping, SET the path not-to-be-cached */
            SetPathNoCache,
            /* rule mapping, SET the path not-to-be-logged */
            SetPathNoLog,
            /* VMS user profile does not apply to this path */
            SetPathNoProfile,
            /* rule mapping, SET the path allowing stream-LF conversion */
            SetPathStmLF,
            /* an "Xray" returns full response as plain-text (diagnostic) */
            Xray;


   int  /* counter representing request activity has be incremented */
        AccountingDone,
        /* could not have come from cache, could have but didn't, did! */
        CacheInvolvement,
        /* multi-purpose, when loading and reading cache */
        CacheContentCount,
        /* length of heap memory pointed to by CgiBufferPtr */
        CgiBufferLength,
        /* remaining space in heap memory pointed to by CgiBufferPtr */
        CgiBufferRemaining,
        /* length of script end-of-file string */
        CgiEofLength,
        /* how many attempts have been made to restart a CGI script */
        CgiScriptRetryCount,
        /* count of bytes received so far in a POSTed or PUTed body */
        ContentCount,
        /* "Content-Length:" value */
        ContentLength,
        /* "Content-Type:" length */
        ContentTypeLength,
        /* size of buffer space for sys$output from DCL subprocess */
        *DclSysOutputSize,
        /* processing request header, count to find first blank line */
        HeaderConsecutiveNewLineCount,
        /* "Accept:" length */
        HttpAcceptLength,
        /* "Accept-Charset:" length */
        HttpAcceptCharsetLength,
        /* "Accept-Encoding:" length */
        HttpAcceptEncodingLength,
        /* "Accept-Language:" length */
        HttpAcceptLangLength,
        /* "Forwarded:" line(s) length */
        HttpForwardedLength,
        /* length of error report text/query-string */
        ErrorReportLength,
        /* value representation of request HTTP method */
        HttpMethod,
        /* value representation of HTTP protocol version number */
        HttpVersion,
        /* length of request header field */
        HttpAuthorizationLength,
        /* length of request header field */
        HttpRefererLength,
        /* length of request header field */
        HttpUserAgentLength,
        /* length of mapped path */
        MappedPathLength,
        /* interge representing prefered message language */
        MsgLanguage,
        /* size of network read buffer */
        NetReadBufferSize,
        /* length of path from HTTP header */
        PathInfoLength,
        /* length of query string from HTTP header */
        QueryStringLength,
        /* lenth of resource (path plus query string) from HTTp header */
        ResourceLength,
        /* number of characters up until the blank line */
        RequestHeaderLength,
        /* characters in the request line (e.g. "GET /path HTTP/1.0\n") */
        RequestLineLength,
        /* count of home page file names tried */
        RequestHomePageIndex,
        /* used during HEAD method to determine when header ends */
        ResponseHeaderNewlineCount,
        /* buffer for data length */
        ResponseHeaderDataLength,
        /* length of reponse header */
        ResponseHeaderLength,
        /* numeric equivalent of the "200", "302", etc., HTTP status code */
        ResponseStatusCode,
        /* VMS status of most recent task */
        TaskStatus;

   unsigned long  /* count of REQUEST-RELATED bytes received from the client */
                  BytesRx,
                  /* count of REQUEST-RELATED bytes sent to the client */
                  BytesTx,
                  /* count of TOTAL bytes received from the client */
                  BytesRawRx,
                  /* count of TOTAL bytes sent to the client */
                  BytesRawTx,
                  /* buffer for hash value used when caching */
                  CacheHashValue,
                  /* response duration */
                  ResponseDuration;

   unsigned long  /* 64 bit VMS internal time, start of request processing */
                  BinaryTime [2],
                  /* 64 bit VMS time from "If-Modified-Since:" header */
                  IfModifiedSinceBinaryTime [2],
                  /* length paramater from "If-Modified-Since:" header */
                  IfModifiedSinceLength;

   unsigned short  /* component equivalent of 'BinaryTime' */
                   NumericTime [7];

   unsigned char  /* multi-purpose, when loading and reading cache */
                  *CacheContentPtr,
                  /* points into cache buffer when loading via record IO */
                  *CacheContentCurrentPtr,
                  /* heap memory in which the CGI variables are generated */
                  *CgiBufferPtr,
                  /* current position in CGI variable buffer */
                  *CgiBufferCurrentPtr,
                  /* script end-of-file string */
                  *CgiEofPtr,
                  /* points to dynamic buffer for body POSTed, PUT, etc. */
                  *ContentBufferPtr,
                  /* used as pointer into above buffer */
                  *ContentPtr,
                  /* buffer space for sys$output from DCL subprocess */
                  *DclSysOutputPtr,
                  /* pointer to other info when generating error report */
                  *ErrorOtherTextPtr,
                  /* pointer to error text/query-string */
                  *ErrorReportPtr,
                  /* pointer to explanation when generating error report */
                  *ErrorTextPtr,
                  /* "Accept:" header lines */
                  *HttpAcceptPtr,
                  /* "Accept-Charset:" header lines */
                  *HttpAcceptCharsetPtr,
                  /* "Accept-Encoding:" header lines */
                  *HttpAcceptEncodingPtr,
                  /* "Accept-Language:" header lines */
                  *HttpAcceptLangPtr,
                  /* pointer to heap storage of "Authorization:" field */
                  *HttpAuthorizationPtr,
                  /* stream's MIME content type */
                  *HttpContentTypePtr,
                  /* "Cookie:" request header encountered */
                  *HttpCookiePtr,
                  /* "Forwarded:" header line */
                  *HttpForwardedPtr,
                  /* "Host:" header line */
                  *HttpHostPtr,
                  /* pointer to heap storage of "If-Modified-Since:" field */
                  *HttpIfModifiedSincePtr,
                  /* "Pragma:" request header encountered */
                  *HttpPragmaPtr,
                  /* "Proxy-Connection:" request header encountered */
                  *HttpProxyConnectionPtr,
                  /* pointer to heap storage of "Referer:" field */
                  *HttpRefererPtr,
                  /* pointer to heap storage of "User-Agent:" field */
                  *HttpUserAgentPtr,
                  /* pointer to heap storage of redirection location path */
                  *LocationPtr,
                  /* pointer to mapped path */
                  *MappedPathPtr,
                  /* pointer to heap storage client-read network buffering */
                  *NetReadBufferPtr,
                  /* pointer to heap storage of request path */
                  *PathInfoPtr,
                  /* pointer to heap storage of request query string */
                  *QueryStringPtr,
                  /* pointer to requested resource (path and query string) */
                  *ResourcePtr,
                  /* pointer to request header */
                  *RequestHeaderPtr,
                  /* pointer to response header */
                  *ResponseHeaderPtr,
                  /* buffer for data pointer */
                  *ResponseHeaderDataPtr,
                  /* rule mapping SET the path with a CGI variable prefix */
                  *SetPathCgiPrefixPtr,
                  /* rule mapping SET the path with a specific charset */
                  *SetPathCharsetPtr,
                  /* rule mapping SET the path with a specific content-type */
                  *SetPathContentTypePtr,
                  /* rule mapping SET the path with a "Index of" format */
                  *SetPathIndexPtr;

   unsigned char  /* GMT-adjusted string equivalent of 'BinaryTime' */
                  GmDateTime [48],
                  /* string reqpresenting the request's HTTP method */
                  HttpMethodName [METHOD_NAME_SIZE],
                  /* authenticated user name from "Authorization:" */
                  RemoteUser [AUTH_MAX_USERNAME_LENGTH+1],
                  /* storage when authenticating remote user */
                  RemoteUserPassword [AUTH_MAX_PASSWORD_LENGTH+1],
                  /* parsed path storage */
                  RequestFileName [256],
                  /* */
                  RequestMappedFile [256],
                  /* */
                  RequestMappedScript [256],
                  /* mapped script name */
                  ScriptName [SCRIPT_NAME_SIZE];

   /* request is being used to fill a cache entry */
   struct CacheStruct  *CachePtr;

   /* AST function pointers */
   void (*ResponseHeaderAstFunctionPtr)(struct RequestStruct*);

   /* storage for maintaining the thread's heap storage */
   struct HeapStruct  *HeapHeadPtr;
   struct HeapStruct  *HeapTailPtr;
   int  HeapSize;

   /* */
   struct ContentTypeStruct  ContentType;

   /* I/O status blocks */
   struct AnIOsb  NetReadIOsb,
                  NetShutdownIOsb,
                  NetWriteIOsb;

   /* RMS structures used for various file processing purposes */
   struct FAB  RequestParseFab;
   struct NAM  RequestParseNam;

   /**************/
   /* task lists */
   /**************/

   /* these two modules can have multiple, concurrent instances */
   struct  ListHeadStruct  DirTaskList,
                           SsiTaskList;

   /* these modules/tasks can only have one total or multiple serially */
   struct  DclTaskStruct      *DclTaskPtr;
   struct  DECnetTaskStruct   *DECnetTaskPtr;
   struct  DescrTaskStruct    *DescrTaskPtr;
   struct  DirTaskStruct      *DirTaskPtr;
   struct  FileTaskStruct     *FileTaskPtr;
   struct  GraphTaskStruct    *GraphTaskPtr;
   struct  HTAdminTaskStruct  *HTAdminTaskPtr;
   struct  IsMapTaskStruct    *IsMapTaskPtr;
   struct  MenuTaskStruct     *MenuTaskPtr;
   struct  ProxyTaskStruct    *ProxyTaskPtr;
   struct  PutTaskStruct      *PutTaskPtr;
   struct  SsiTaskStruct      *SsiTaskPtr;
   struct  UpdTaskStruct      *UpdTaskPtr;

   /****************************************/
   /* storage specific to output buffering */
   /****************************************/

   /* the head of the output buffer list */
   struct ListHeadStruct  OutputBufferList;

   /* pointer to heap storage structure for output buffering */
   struct ListEntryStruct  *OutputBufferStructPtr,
                           *OutputBufferNowStructPtr;

   /* pointer to any AST function */
   void  *OutputBufferAstFunctionPtr;

   boolean  /* all output buffers must be written to network */
            OutputBufferFlush;

   int  /* count of all data in all output buffers */
        OutputBufferCount,
        /* bytes still available in the current output buffer */
        OutputBufferRemaining;

   unsigned char  /* pointer to the start of the current buffer */
                  *OutputBufferPtr,
                  /* pointer used to point into 'OutputBuffer' storage */
                  *OutputBufferCurrentPtr;

   /****************************************************/
   /* storage specific to authentication/authorization */
   /****************************************************/

   boolean  /* user may change SYSUAF password */
            AuthCanChangeSysUafPwd,
            /* authenticated user may only access via https: (SSL) */
            AuthHttpsOnly,
            /* the authentication realm name is a synonym for the SYSUAF */
            AuthRealmIsVmsSynonym,
            /* the user was authenticated from the SYSUAF */
            AuthSysUafAuthenticated;

   unsigned short  /* length of the user profile pointed to */
                   AuthVmsUserProfileLength;

   unsigned long  /* value representation of scheme being used for challenge */
                  AuthChallengeScheme,
                  /* the capability flags of the authorization path */
                  AuthGroupCan,
                  /* VMS identifier value controlling read access */
                  AuthGroupReadVmsIdentifier,
                  /* VMS identifier value controlling write access */
                  AuthGroupWriteVmsIdentifier,
                  /* VMS identifier value controlling authentication */
                  AuthRealmVmsIdentifier,
                  /* the capability flags of the request (ACT ON THESE!) */
                  AuthRequestCan,
                  /* value representation of scheme being used for request */
                  AuthScheme,
                  /* source of read-only group information */
                  AuthSourceGroupRead,
                  /* source of full-access group information */
                  AuthSourceGroupWrite,
                  /* source of realm information */
                  AuthSourceRealm,
                  /* the capability flags of the remote user */
                  AuthUserCan,
                  /* what everybody else can do on that path */
                  AuthWorldCan;

   unsigned long  /* pointer to array of VMS identifiers (zero-terminated) */
                  *AuthVmsIdentifiersPtr;

   unsigned char  /* BASIC challenge response header field */
                  *AuthBasicChallengePtr,
                  /* DIGEST challenge response header field */
                  *AuthDigestChallengePtr,
                  /* when digest authenticating, the client-supplied nonce */
                  *AuthDigestNoncePtr,
                  /* when digest authenticating, the client-supplied digest */
                  *AuthDigestResponsePtr,
                  /* when digest authenticating, the client-supplied URI */
                  *AuthDigestUriPtr,
                  /* pointer to authorization group name */
                  *AuthGroupReadPtr,
                  /* pointer to string of IP-address/username masks */
                  *AuthGroupRestrictListPtr,
                  /* pointer to authorization group name */
                  *AuthGroupWritePtr,
                  /* pointer to authorization realm name */
                  *AuthRealmPtr,
                  /* pointer to authorization realm synonym string */
                  *AuthRealmDescriptionPtr,
                  /* user profile is used by sys$check_access() */
                  *AuthVmsUserProfilePtr,
                  /* pointer to string of IP-address/username masks */
                  *AuthWorldRestrictListPtr;

   char  /* basic, digest, etc. */
         AuthType [16];


   /***************/
   /* zeroed end! */
   /***************/

   unsigned long  ZeroedEnd;
};

#ifdef __ALPHA
#   pragma member_alignment __restore
#endif

#endif /* WASD_H_LOADED */

/*****************************************************************************/
